/**
 * @file BackUpState.h
 * @brief Backup sync states
 *
 * (c) 2023- by Mega Limited, Auckland, New Zealand
 *
 * This file is part of the MEGA SDK - Client Access Engine.
 *
 * Applications using the MEGA API must present a valid application key
 * and comply with the the rules set forth in the Terms of Service.
 *
 * The MEGA SDK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * @copyright Simplified (2-clause) BSD License.
 *
 * You should have received a copy of the license along with this
 * program.
 */

typedef NS_ENUM(NSInteger, BackUpState) {
    BackUpStateInvalid = -1,
    BackUpStateNotInitialized = 0,
    BackUpStateActive = 1,
    BackUpStateFailed = 2,
    BackUpStateTemporaryDisabled = 3,
    BackUpStateDisabled = 4,
    BackUpStatePauseUp = 5,
    BackUpStatePauseDown = 6,
    BackUpStatePauseFull = 7,
    BackUpStateDeleted = 8,
    BackUpStateUnknown = 9
};
