/**
 * @file mega/posix/megaconsolewaiter.h
 * @brief Posix event/timeout handling, listens for stdin input
 *
 * (c) 2013-2014 by Mega Limited, Wellsford, New Zealand
 *
 * This file is part of the MEGA SDK - Client Access Engine.
 *
 * Applications using the MEGA API must present a valid application key
 * and comply with the the rules set forth in the Terms of Service.
 *
 * The MEGA SDK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * @copyright Simplified (2-clause) BSD License.
 *
 * You should have received a copy of the license along with this
 * program.
 */

#ifndef CONSOLE_WAIT_CLASS
#define CONSOLE_WAIT_CLASS PosixConsoleWaiter

#include "megawaiter.h"

namespace mega {
struct PosixConsoleWaiter : public PosixWaiter
{
    int wait();
};
} // namespace

#endif
